@@ -1,9 +1,9 @@
 from contextlib  import contextmanager
-from version import __VERSION__
+import operator
 from PyQt4 import QtGui, QtCore, QtSql
-from utils import showWaitCursorDecorator, showWaitCursor
-from models import RecordModel, SortProxyModel, CheckComboModel, TagModel
-from csvdecoder import Decoder, DecoderException
+from version import __VERSION__
+import utils
+from models import RecordModel, RecordProxyModel, AccountsModel, TagModel, TagProxyModel
 from database import db
 from ui_pydosh import Ui_pydosh
 from dialogs import SettingsDialog, ImportDialog
@@ -39,21 +39,6 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 		self.removeTagButton.setEnabled(False)
 
 		self.accountCombo.setDefaultText('all')
-		self.accountCombo.selectionChanged.connect(self.setFilter)
-		self.checkedCombo.currentIndexChanged.connect(self.setFilter)
-		self.tagsCombo.currentIndexChanged.connect(self.setFilter)
-		self.inoutCombo.currentIndexChanged.connect(self.setFilter)
-		self.descEdit.textChanged.connect(self.setFilter)
-		self.amountEdit.textChanged.connect(self.setFilter)
-		self.amountEdit.controlKeyPressed.connect(self.controlKeyPressed)
-		self.toggleCheckButton.clicked.connect(self.toggleSelected)
-		self.deleteButton.clicked.connect(self.deleteRecords)
-		self.dateCombo.currentIndexChanged.connect(self.dateRangeSelected)
-		self.startDateEdit.dateChanged.connect(self.setFilter)
-		self.endDateEdit.dateChanged.connect(self.setFilter)
-		self.reloadButton.clicked.connect(self.reset)
-		self.addTagButton.clicked.connect(self.addTag)
-		self.removeTagButton.clicked.connect(self.removeTag)
 
 		self.connectionStatusText.setText('connected to %s@%s' % (db.database, db.hostname))
 
@@ -65,96 +50,176 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 
 		self.addActions()
 
-		self.inTotalLabel.setFrameStyle(QtGui.QFrame.StyledPanel | QtGui.QFrame.Sunken)
-		self.outTotalLabel.setFrameStyle(QtGui.QFrame.StyledPanel | QtGui.QFrame.Sunken)
-		self.recordCountLabel.setFrameStyle(QtGui.QFrame.StyledPanel | QtGui.QFrame.Sunken)
+		styleSheet = 'QLabel {font: bold;}'
+		self.inTotalLabel.setStyleSheet(styleSheet)
+		self.outTotalLabel.setStyleSheet(styleSheet)
+		self.recordCountLabel.setStyleSheet(styleSheet)
 
-		self.__signalsToBlock = (
-				self.accountCombo,
-				self.checkedCombo,
-				self.tagsCombo,
-				self.inoutCombo,
-				self.descEdit,
-				self.amountEdit,
-				self.dateCombo,
-				self.startDateEdit,
-				self.endDateEdit,
-		)
+		# Date ranges
+		self.maxInsertDate = None
+		self.populateDates()
 
-		self.filterTagIds = set()
+		# Set up record model
+		recordModel = RecordModel(self)
+		recordModel.setTable('records')
+		recordModel.setEditStrategy(QtSql.QSqlTableModel.OnFieldChange)
+		recordProxyModel = RecordProxyModel(self)
+		recordProxyModel.setDynamicSortFilter(True)
+		recordProxyModel.setSourceModel(recordModel)
+		recordProxyModel.sort(enum.kRecordColumn_Date, QtCore.Qt.AscendingOrder)
+		self.tableView.setModel(recordProxyModel)
+		recordModel.select()
 
-		with self.blockAllSignals():
+		# Set up record view
+		self.tableView.verticalHeader().hide()
+		self.tableView.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
+		self.tableView.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
+		self.tableView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
+		self.tableView.setColumnHidden(enum.kRecordColumn_RecordId, True)
+		self.tableView.setColumnHidden(enum.kRecordColumn_AccountTypeId, True)
+		self.tableView.setColumnHidden(enum.kRecordColumn_CheckDate, True)
+		self.tableView.setColumnHidden(enum.kRecordColumn_RawData, True)
+		self.tableView.setColumnHidden(enum.kRecordColumn_InsertDate, True)
+		self.tableView.setSortingEnabled(True)
+		self.tableView.horizontalHeader().setResizeMode(enum.kRecordColumn_Description, QtGui.QHeaderView.Stretch)
+		self.tableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
 
-			model = RecordModel(self)
-			model.setTable('records')
-			model.setEditStrategy(QtSql.QSqlTableModel.OnFieldChange)
-			model.dataChanged.connect(self.setFilter)
-			model.select()
+		# Set up tag model
+		tagModel = TagModel(self)
+		tagProxyModel = TagProxyModel(self)
+		tagProxyModel.setSourceModel(tagModel)
+		tagProxyModel.sort(enum.kTagsColumn_Amount_out, QtCore.Qt.AscendingOrder)
+		self.tagView.setModel(tagProxyModel)
 
-			proxyModel = SortProxyModel(self)
-			proxyModel.setSourceModel(model)
-			proxyModel.sort(enum.kRecordColumn_Date, QtCore.Qt.AscendingOrder)
+		# Set up tag view
+		self.tagView.verticalHeader().hide()
+		self.tagView.setColumnHidden(enum.kTagsColumn_TagId, True)
+		self.tagView.setColumnHidden(enum.kTagsColumn_RecordIds, True)
+		self.tagView.setSortingEnabled(True)
+		self.tagView.sortByColumn(enum.kTagsColumn_TagName, QtCore.Qt.AscendingOrder)
+		self.tagView.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
+		self.tagView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
+		self.tagView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
+		self.tagView.setStyleSheet('QTableView {background-color: transparent;}')
+		self.tagView.setShowGrid(False)
 
-			self.tableView.setModel(proxyModel)
-			self.tableView.verticalHeader().hide()
-			self.tableView.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
-			self.tableView.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
-			self.tableView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
+		# Set up account model
+		accountModel = AccountsModel(self)
+		accountModel.setTable('accounttypes')
+		accountModel.setFilter("""
+			accounttypeid IN (
+				SELECT distinct accounttypeid
+				FROM records
+				WHERE userid=%d)
+			""" % db.userId)
+		accountModel.select()
+		self.accountCombo.setModelColumn(enum.kAccountTypeColumn_AccountName)
+		self.accountCombo.setModel(accountModel)
 
-			self.tableView.setColumnHidden(enum.kRecordColumn_RecordId, True)
-			self.tableView.setColumnHidden(enum.kRecordColumn_CheckDate, True)
-			self.tableView.setColumnHidden(enum.kRecordColumn_RawData, True)
-			self.tableView.setColumnHidden(enum.kRecordColumn_InsertDate, True)
-			self.tableView.setSortingEnabled(True)
-
-			self.tableView.horizontalHeader().setResizeMode(enum.kRecordColumn_Description, QtGui.QHeaderView.Stretch)
-			self.tableView.selectionModel().selectionChanged.connect(self.recordSelectionChanged)
-
-			self.tableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
-			self.tableView.customContextMenuRequested.connect(self.tagEditPopup)
-
-			model = TagModel(self)
-			model.tagsChanged.connect(self.tagModelChanged)
-			model.selectionChanged.connect(self.setTagFilter)
-			proxyModel = QtGui.QSortFilterProxyModel(self)
-			proxyModel.setSourceModel(model)
-			proxyModel.sort(enum.kTagsColumn_Amount_out, QtCore.Qt.AscendingOrder)
-
-			self.tagView.setModel(proxyModel)
-			self.tagView.setColumnHidden(enum.kTagsColumn_TagId, True)
-			self.tagView.setColumnHidden(enum.kTagsColumn_RecordIds, True)
-			self.tagView.setSortingEnabled(True)
-			self.tagView.sortByColumn(enum.kTagsColumn_TagName, QtCore.Qt.AscendingOrder)
-			self.tagView.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
-			self.tagView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
-			self.tagView.selectionModel().selectionChanged.connect(self.enableRemoveTagButton)
-
-			self.tagView.verticalHeader().hide()
-			self.tagView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
-			self.tagView.setStyleSheet('QTableView {background-color: transparent;}')
-			self.tagView.setShowGrid(False)
-
-			# Set sql data model for account types
-			self.accountCombo.clear()
-
-			accountModel = CheckComboModel(self)
-			accountModel.setTable('accounttypes')
-			accountModel.setFilter("""
-				accounttypeid IN (
-					SELECT distinct accounttypeid
-					FROM records
-					WHERE userid=%d)
-				""" % db.userId)
-
-			accountModel.select()
-			accountModel.setUserRoleColumn(enum.kAccountTypeColumn_AccountTypeId)
-			self.accountCombo.setModelColumn(enum.kAccountTypeColumn_AccountName)
-			self.accountCombo.setModel(accountModel)
+		#
+		# Set up connections
+		#
+		self.scrolltoEdit.textChanged.connect(self.scrollTo)
+		self.toggleCheckButton.clicked.connect(self.toggleSelected)
+		self.deleteButton.clicked.connect(self.deleteRecords)
+		self.dateCombo.currentIndexChanged.connect(self.setDateRange)
+		self.reloadButton.clicked.connect(self.reset)
+		self.addTagButton.clicked.connect(self.addTag)
+		self.removeTagButton.clicked.connect(self.removeTag)
+		recordModel.dataChanged.connect(self.updateTagFilter)
+		recordProxyModel.filterChanged.connect(self.updateTagFilter)
+		recordProxyModel.modelReset.connect(self.updateTagFilter)
+		self.tableView.selectionModel().selectionChanged.connect(self.recordSelectionChanged)
+		self.tableView.customContextMenuRequested.connect(self.tagEditPopup)
+		tagModel.tagsChanged.connect(self.tagModelChanged)
+		tagModel.selectionChanged.connect(recordProxyModel.setTagFilter)
+		self.tagView.selectionModel().selectionChanged.connect(self.enableRemoveTagButton)
+		self.startDateEdit.dateChanged.connect(recordProxyModel.setStartDate)
+		self.endDateEdit.dateChanged.connect(self.endDateChanged)
+		self.endDateEdit.dateChanged.connect(recordProxyModel.setEndDate)
+		self.accountCombo.selectionChanged.connect(self.accountSelectionChanged)
+		self.tagsCombo.currentIndexChanged.connect(self.tagSelectionChanged)
+		self.checkedCombo.currentIndexChanged.connect(self.checkedSelectionChanged)
+		self.inoutCombo.currentIndexChanged.connect(self.inOutSelectionChanged)
+		self.descEdit.textChanged.connect(recordProxyModel.setDescriptionFilter)
+		self.amountEdit.controlKeyPressed.connect(self.controlKeyPressed)
+		self.amountEdit.textChanged.connect(self.amountFilterChanged)
 
 		self.reset()
 
-	def showAbout(self):
+	def accountSelectionChanged(self, indexes):
+		""" Tell the proxy filter to filter on selected account ids
+		"""
+		model = self.accountCombo.model()
+		accountIds = []
+		for index in indexes:
+			accountIds.append(model.index(index.row(), enum.kAccountTypeColumn_AccountTypeId).data().toPyObject())
 
+		self.tableView.model().setAccountFilter(accountIds)
+
+	def tagSelectionChanged(self, index):
+		""" Basic tag filter (has tags or not)
+		"""
+		state = self.tagsCombo.itemData(index, QtCore.Qt.UserRole).toPyObject()
+		if state == enum.kTagCombo_With:
+			value = True
+		elif state == enum.kTagCombo_Without:
+			value = False
+		else:
+			value = None
+		self.tableView.model().setHasTagsFilter(value)
+
+	def checkedSelectionChanged(self, index):
+		""" Set filter for checked records
+		"""
+		state = self.checkedCombo.itemData(index, QtCore.Qt.UserRole).toPyObject()
+		if state == enum.kCheckedStatus_Checked:
+			value = True
+		elif state == enum.kCheckedStatus_UnChecked:
+			value = False
+		else:
+			value = None
+		self.tableView.model().setCheckedFilter(value)
+
+	def inOutSelectionChanged(self, index):
+		""" Set filter for in/out amount
+		"""
+		state = self.inoutCombo.itemData(index, QtCore.Qt.UserRole).toPyObject()
+		if state == enum.kInOutStatus_In:
+			value = True
+		elif state == enum.kInOutStatus_Out:
+			value = False
+		else:
+			value = None
+		self.tableView.model().setCreditFilter(value)
+
+	def amountFilterChanged(self, text):
+		if text:
+			if text.contains(QtCore.QRegExp('[<>=]+')):
+				# looks like we have operators, test validity and amount
+				operatorMap = {
+					'=':  operator.eq,
+					'>':  operator.gt,
+					'<':  operator.lt,
+					'>=': operator.ge,
+					'<=': operator.le
+				}
+				rx = QtCore.QRegExp('^(=|>|<|>=|<=)([\\.\\d+]+)')
+				if rx.indexIn(text) != -1 and str(rx.cap(1)) in operatorMap:
+					self.tableView.model().setAmountFilter(rx.cap(2), operatorMap[str(rx.cap(1))])
+				else:
+					# Input not complete yet.
+					return
+			else:
+				# No operator supplied - treat amount as a string
+				self.tableView.model().setAmountFilter(text)
+		else:
+			# No filter
+			self.tableView.model().setAmountFilter(None)
+
+	def showAbout(self):
+		""" Show 'about' info
+		"""
 		QtGui.QMessageBox.about(self,
 			'About pydosh',
 			"<html><p><h2>pydosh</h2></p>"
@@ -166,55 +231,76 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 			"enjoy!</html>" % (__VERSION__, QtCore.QT_VERSION_STR)
 			)
 
-	def selectDateRange(self):
-		selected = self.dateCombo.itemData(self.dateCombo.currentIndex(), QtCore.Qt.UserRole).toPyObject()
+	def endDateChanged(self, date):
+		""" End date has changed - if date mode is "previous" month, 
+			the update start date accordingly
+		"""
+		if self.dateCombo.currentIndex() == enum.kDate_PreviousMonth:
+			self.startDateEdit.setDate(self.endDateEdit.date().addMonths(-1))
+
+	@utils.showWaitCursorDecorator
+	def setDateRange(self, index=None):
+		""" Set date fields according to date combo selection mode
+		"""
+		if index is None:
+			index = self.dateCombo.currentIndex()
+
+		proxyModel = self.tableView.model()
+
+		if index == enum.kDate_All:
+			self.startDateEdit.setEnabled(True)
+			self.endDateEdit.setEnabled(True)
 
-		if selected == enum.kDate_All:
 			self.startDateEdit.setDate(self.startDateEdit.minimumDate())
 			self.endDateEdit.setDate(self.endDateEdit.maximumDate())
-			self.startDateEdit.setEnabled(True)
-			self.endDateEdit.setEnabled(True)
-		elif selected == enum.kDate_PreviousMonth:
-			self.startDateEdit.setDate(self.endDateEdit.date().addMonths(-1))
+
+		elif index == enum.kDate_PreviousMonth:
 			self.endDateEdit.setEnabled(True)
 			self.startDateEdit.setEnabled(False)
-		elif selected == enum.kDate_PreviousYear:
-			self.startDateEdit.setDate(self.endDateEdit.date().addYears(-1))
+
+			self.endDateEdit.setDate(self.endDateEdit.maximumDate())
+			self.startDateEdit.setDate(self.endDateEdit.date().addMonths(-1))
+
+		elif index == enum.kDate_PreviousYear:
 			self.startDateEdit.setEnabled(True)
 			self.endDateEdit.setEnabled(True)
-		elif selected == enum.kdate_LastImport:
+
+			self.endDateEdit.setDate(self.endDateEdit.maximumDate())
+			self.startDateEdit.setDate(self.endDateEdit.date().addYears(-1))
+
+		elif index == enum.kdate_LastImport:
 			self.startDateEdit.setEnabled(False)
 			self.endDateEdit.setEnabled(False)
 
+			proxyModel.setInsertDate(insertDate=self.maxInsertDate)
+
+			startDate = QtCore.QDate()
+			endDate = QtCore.QDate()
+
+			# Find the min/max dates for the last import from the proxy model 
+			for row in xrange(proxyModel.rowCount()):
+				dateForRow = proxyModel.index(row, enum.kRecordColumn_Date).data().toDate()
+				if not endDate.isValid() or dateForRow > endDate:
+					endDate = dateForRow
+				if not startDate.isValid() or dateForRow < startDate:
+					startDate = dateForRow
+
+			# This is for information only, so don't trigger a query on the proxy
+			with utils.signalsBlocked(self.endDateEdit, self.startDateEdit):
+				self.startDateEdit.setDate(startDate)
+				self.endDateEdit.setDate(endDate)
+
 	def settingsDialog(self):
 		""" Launch the settings dialog widget to configure account information
 		"""
 		dialog = SettingsDialog(self)
 		dialog.exec_()
-		self.setFilter()
 
 	def importDialog(self):
 		""" Launch the import dialog
 		"""
-		combo = QtGui.QComboBox(self)
-		model = QtSql.QSqlTableModel(self)
-		model.setTable('accounttypes')
-		model.select()
-		combo.setModel(model)
-		combo.setModelColumn(enum.kAccountTypeColumn_AccountName)
-
-		# find the previous accounttype that was used (if any)
+		# try and set last import directory
 		settings = QtCore.QSettings()
-		accounttype = settings.value("options/importaccounttype").toString()
-
-		index = combo.findText(accounttype)
-
-		if index != -1:
-			combo.setCurrentIndex(index)
-		else:
-			combo.setCurrentIndex(0)
-
-		label = QtGui.QLabel('Account type:')
 		importDir = settings.value("options/importdirectory").toString()
 
 		if importDir.isEmpty():
@@ -222,54 +308,23 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 
 		dialog = QtGui.QFileDialog(self, 'Open File', importDir, "*.csv")
 		dialog.setFileMode(QtGui.QFileDialog.ExistingFiles)
-		dialog.setOption(QtGui.QFileDialog.DontUseNativeDialog)
-
-		gridbox = dialog.layout()
-
-		if gridbox:
-			gridbox.addWidget(label)
-			gridbox.addWidget(combo)
 
 		if not dialog.exec_():
 			return
 
-		accountId, ok = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_AccountTypeId).data().toInt()
-		if not ok:
-			QtGui.QMessageBox.critical(self, 'Import Error', 'No account type specified', QtGui.QMessageBox.Ok)
-			return
-
 		fileNames = QtCore.QStringList([QtCore.QFileInfo(f).fileName() for f in dialog.selectedFiles()])
-		dateField = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_DateField).data()
-		descriptionField = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_DescriptionField).data()
-		creditField = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_CreditField).data()
-		debitField = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_DebitField).data()
-		currencySign = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_CurrencySign).data()
-		dateFormat = combo.model().index(combo.currentIndex(), enum.kAccountTypeColumn_DateFormat).data()
 
 		# Save the settings for next time
-		settings.setValue('options/importaccounttype', combo.currentText())
 		settings.setValue('options/importdirectory', dialog.directory().absolutePath())
 
-		with showWaitCursor():
-			try:
-				decoder = Decoder(
-						dateField,
-						descriptionField,
-						creditField,
-						debitField,
-						currencySign,
-						dateFormat,
-						dialog.selectedFiles())
-			except DecoderException, exc:
-				QtGui.QMessageBox.critical(self, 'Import Error', str(exc), QtGui.QMessageBox.Ok)
-				return
-
-			dialog = ImportDialog(decoder.records, accountId, self)
-			dialog.setWindowTitle(fileNames.join(', '))
+		dialog = ImportDialog(dialog.selectedFiles(), self)
+		dialog.setWindowTitle(fileNames.join(', '))
 
 		if dialog.exec_():
-			self.accountCombo.model().select()
-			self.reset()
+			self.tableView.model().sourceModel().select()
+			self.accountCombo.reset()
+			self.populateDates()
+			self.setDateRange()
 
 	def recordSelectionChanged(self):
 		""" Only enable buttons if a selection has been made
@@ -322,23 +377,16 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 				QtGui.QMessageBox.Yes|QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
 			return
 
-		with showWaitCursor():
+		with utils.showWaitCursor():
 			indexes = [proxyModel.mapToSource(index) for index in selectionModel.selectedRows()]
-			if proxyModel.sourceModel().deleteRecords(indexes):
-				self.accountCombo.model().select()
-			else:
+			if not proxyModel.sourceModel().deleteRecords(indexes):
 				QtGui.QMessageBox.critical(self, 'Database Error',
 					proxyModel.sourceModel().lastError().text(), QtGui.QMessageBox.Ok)
-
-	@contextmanager
-	def blockAllSignals(self):
-		try:
-			for widget in self.__signalsToBlock:
-				widget.blockSignals(True)
-			yield
-		finally:
-			for widget in self.__signalsToBlock:
-				widget.blockSignals(False)
+			
+			# Finally, re-populate accounts and date range in case this has changed
+			self.accountCombo.reset()
+			self.populateDates()
+			self.setDateRange()
 
 	@contextmanager
 	def keepSelection(self):
@@ -366,69 +414,105 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 					self.tableView.selectRow(proxyModel.mapFromSource(match[0]).row())
 
 			self.tableView.setSelectionMode(selectionMode)
+			
+			selected = self.tableView.selectionModel().selectedRows()
+
+			# Scroll to first selected row
+			if selected:
+				self.tableView.scrollTo(selected[0], QtGui.QAbstractItemView.EnsureVisible)
 
 	def populateDates(self):
+		""" Set date fields to min and max values
+		"""
 		query = QtSql.QSqlQuery("""
-				SELECT MIN(date), MAX(date)
-				FROM records
-				WHERE userid=%d
+			SELECT MIN(date), 
+				   MAX(date),  
+				   MAX(insertdate)
+			  FROM records
+			 WHERE userid=%d
 			""" % db.userId)
 
 		if query.next():
 			startDate = query.value(0).toDate()
 			endDate = query.value(1).toDate()
+			maxInsertDateTime = query.value(2).toDateTime()
 
 			self.startDateEdit.setDateRange(startDate, endDate)
 			self.endDateEdit.setDateRange(startDate, endDate)
+			self.maxInsertDate = maxInsertDateTime
 
-			self.startDateEdit.setDate(startDate)
-			self.endDateEdit.setDate(endDate)
-
-	@showWaitCursorDecorator
+	@utils.showWaitCursorDecorator
 	def toggleSelected(self, *args):
 
 		""" Toggle checked status on all selected rows in view.
 		"""
-		# Get recordids from all selected rows
 		selectionModel = self.tableView.selectionModel()
-		if selectionModel is None:
-			return
-
 		proxyModel = self.tableView.model()
-		indexes = [proxyModel.mapToSource(proxyIndex) for proxyIndex in selectionModel.selectedRows()]
-		if not proxyModel.sourceModel().setItemsChecked(indexes):
-			QtGui.QMessageBox.critical(self, 'Database Error', 
-					proxyModel.sourceModel().lastError().text(), QtGui.QMessageBox.Ok)
 
-	def reset(self):
-		if self.tableView.model() is None:
-			return
+		editStrategy = proxyModel.sourceModel().editStrategy()
+		proxyModel.sourceModel().setEditStrategy(QtSql.QSqlTableModel.OnManualSubmit)
+
+		with utils.signalsBlocked(proxyModel.sourceModel()):
+			for proxyIndex in selectionModel.selectedRows():
+				# Need the index of checked column
+				proxyIndex = proxyModel.index(proxyIndex.row(), enum.kRecordColumn_Checked)
+				index = proxyModel.mapToSource(proxyIndex)
+
+				if index.data(QtCore.Qt.CheckStateRole).toPyObject() == QtCore.Qt.Checked:
+					proxyModel.sourceModel().setData(index, QtCore.QVariant(QtCore.Qt.Unchecked), QtCore.Qt.CheckStateRole)
+				elif index.data(QtCore.Qt.CheckStateRole).toPyObject() == QtCore.Qt.Unchecked:
+					proxyModel.sourceModel().setData(index, QtCore.QVariant(QtCore.Qt.Checked), QtCore.Qt.CheckStateRole)
+		
+			proxyModel.sourceModel().submitAll()
+
+		proxyModel.sourceModel().setEditStrategy(editStrategy)
+
+		with self.keepSelection():
+			self.tableView.model().sourceModel().reset()
+
+	@utils.showWaitCursorDecorator
+	def reset(self, *args):
+		""" Reset all filters and combo boxes to a default state
+		"""
+		signalsToBlock = (
+				self.accountCombo,
+				self.checkedCombo,
+				self.tagsCombo,
+				self.inoutCombo,
+				self.descEdit,
+				self.scrolltoEdit,
+				self.amountEdit,
+				self.dateCombo,
+				self.startDateEdit,
+				self.endDateEdit,
+				self.tableView.model(),
+		)
+
+		with utils.signalsBlocked(signalsToBlock):
+			self.tagView.model().sourceModel().clearSelection()
+			self.tableView.model().clearFilters()
+			
+			self.dateCombo.setCurrentIndex(enum.kDate_PreviousMonth)
+
+			# Signals blocked so need to reset filter manually
+			self.setDateRange()
+			self.tableView.model().setStartDate(self.startDateEdit.date())
+			self.tableView.model().setEndDate(self.endDateEdit.date())
 
-		with self.blockAllSignals():
-			self.populateDates()
 			self.checkedCombo.setCurrentIndex(enum.kCheckedStatus_All)
 			self.tagsCombo.setCurrentIndex(enum.kCheckedStatus_All)
-			self.dateCombo.setCurrentIndex(enum.kDate_PreviousYear)
-			self.selectDateRange()
-			self.accountCombo.clearAll()
 			self.inoutCombo.setCurrentIndex(enum.kInOutStatus_All)
+			self.accountCombo.clearAll()
 			self.amountEdit.clear()
 			self.descEdit.clear()
-			self.endDateEdit.setEnabled(True)
 			self.tableView.sortByColumn(enum.kRecordColumn_Date, QtCore.Qt.DescendingOrder)
 
-		# Reset tagView - block signals as we're calling setFilter anyway
-		tagModel = self.tagView.model().sourceModel()
-		tagModel.blockSignals(True)
-		tagModel.clearSelection()
-		tagModel.blockSignals(False)
-		self.setFilter()
+		# Need signals to clear highlight filter on model
+		self.scrolltoEdit.clear()
+		self.tableView.model().reset()
 
-	def dateRangeSelected(self):
-		""" Date combo has been changed. Set the date fields and refresh the records model
-		"""
-		self.selectDateRange()
-		self.setFilter()
+		self.tagView.resizeColumnsToContents()
+		self.tableView.resizeColumnsToContents()
 
 	def addActions(self):
 		quitAction = QtGui.QAction(QtGui.QIcon(':/icons/exit.png'), '&Quit', self)
@@ -464,121 +548,21 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 		helpMenu = self.menuBar().addMenu('&Help')
 		helpMenu.addAction(aboutAction)
 
-
 	def displayRecordCount(self):
-		numRecords = 0
-		totalRecords = 0
 		inTotal = 0.0
 		outTotal = 0.0
+
 		model = self.tableView.model()
-
-		if model:
-			numRecords = model.rowCount()
-
-			query = QtSql.QSqlQuery('SELECT COUNT(*) FROM records WHERE userid=%d' % db.userId)
-			query.next()
-			totalRecords = query.value(0).toPyObject()
-
-			for i in xrange(numRecords):
-				amount, _ = model.sourceModel().record(i).value(enum.kRecordColumn_Amount).toDouble()
-				if amount > 0.0:
-					inTotal += amount
-				else:
-					outTotal += abs(amount)
+		for row in xrange(model.rowCount()):
+			amount = model.index(row, enum.kRecordColumn_Amount).data(QtCore.Qt.UserRole).toPyObject()
+			if amount > 0.0:
+				inTotal += amount
+			else:
+				outTotal += abs(amount)
 
 		self.inTotalLabel.setText(QtCore.QString("%L1").arg(inTotal, 0, 'f', 2))
 		self.outTotalLabel.setText(QtCore.QString("%L1").arg(outTotal, 0, 'f', 2))
-		self.recordCountLabel.setText('%d / %d' % (numRecords, totalRecords))
-
-	@showWaitCursorDecorator
-	def setFilter(self, *args):
-		model = self.tableView.model()
-
-		if model is None:
-			return
-
-		queryFilter = []
-
-		# Account filter
-		accountIds = [index.data(QtCore.Qt.UserRole).toPyObject() for index in self.accountCombo.checkedIndexes()]
-		if accountIds:
-			queryFilter.append('r.accounttypeid in (%s)' % ', '.join(str(acid) for acid in accountIds))
-
-		# Date filter
-		if self.dateCombo.itemData(self.dateCombo.currentIndex(), QtCore.Qt.UserRole).toPyObject() == enum.kdate_LastImport:
-			queryFilter.append("""
-				r.insertdate = (
-					SELECT MAX(insertdate)
-					FROM records)
-			""")
-		else:
-			startDate = self.startDateEdit.date()
-			endDate = self.endDateEdit.date()
-
-			if startDate.isValid() and endDate.isValid():
-				queryFilter.append("r.date >= '%s'" % startDate.toString(QtCore.Qt.ISODate))
-				queryFilter.append("r.date <= '%s'" % endDate.toString(QtCore.Qt.ISODate))
-
-		# Basic tag filter
-		state = self.tagsCombo.itemData(self.tagsCombo.currentIndex(), QtCore.Qt.UserRole).toPyObject()
-		if state == enum.kTagCombo_With:
-			queryFilter.append('t.tagname is not null')
-		elif state == enum.kTagCombo_Without:
-			queryFilter.append('t.tagname is null')
-
-		# checked state filter
-		state = self.checkedCombo.itemData(self.checkedCombo.currentIndex(), QtCore.Qt.UserRole).toPyObject()
-		if state == enum.kCheckedStatus_Checked:
-			queryFilter.append('r.checked=1')
-		elif state == enum.kCheckedStatus_UnChecked:
-			queryFilter.append('r.checked=0')
-
-		# money in/out filter
-		state = self.inoutCombo.itemData(self.inoutCombo.currentIndex(), QtCore.Qt.UserRole).toPyObject()
-		if state == enum.kInOutStatus_In:
-			queryFilter.append('r.amount > 0')
-		elif state == enum.kInOutStatus_Out:
-			queryFilter.append('r.amount < 0')
-
-		# description filter
-		if self.descEdit.text():
-			queryFilter.append("lower(r.description) LIKE '%%%s%%'" % self.descEdit.text().toLower())
-
-		# amount filter. May contain operators < > <= or >=
-		amountFilter = self.amountEdit.text()
-		if amountFilter:
-			if amountFilter.contains(QtCore.QRegExp('[<>=]+')):
-				# looks like we have operators, test validity and amount
-				rx = QtCore.QRegExp('^(=|>|<|>=|<=)([\\.\\d+]+)')
-				if rx.indexIn(amountFilter) != -1:
-					queryFilter.append('abs(r.amount) %s %s' % (rx.cap(1), rx.cap(2)))
-				else:
-					# Input not complete yet.
-					return
-			else:
-				# No operator supplied - treat amount as a string
-				queryFilter.append(
-					"(CAST(r.amount AS char(10)) LIKE '%s%%' OR CAST(r.amount AS char(10)) LIKE '-%s%%')" %
-					(amountFilter, amountFilter))
-
-		if self.filterTagIds:
-			queryFilter.append("""
-				r.recordid IN (
-					SELECT recordid
-					FROM recordtags
-					WHERE tagid in (%s))
-				""" % ', '.join([str(tagid) for tagid in self.filterTagIds]))
-
-		with self.keepSelection():
-			model.sourceModel().setFilter('\nAND '.join(queryFilter))
-		#print model.sourceModel().query().lastQuery().replace(' AND ', '').replace('\n', ' ')
-
-		self.updateTagFilter()
-
-		self.tagView.updateGeometry()
-		self.tableView.resizeColumnsToContents()
-		self.tagView.resizeColumnsToContents()
-		self.displayRecordCount()
+		self.recordCountLabel.setText('%d / %d' % (model.rowCount(), self.tableView.model().sourceModel().rowCount()))
 
 	def tagModelChanged(self):
 		""" Tag model has changed - call select on record model to refresh the changes (in tag column)
@@ -587,16 +571,19 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 		with self.keepSelection():
 			self.tableView.model().sourceModel().select()
 
-	def updateTagFilter(self):
+	@utils.showWaitCursorDecorator
+	def updateTagFilter(self, *args):
 		""" Tell the tag model to limit tag amounts to current displayed records
 		"""
 		recordIds = []
-		model = self.tableView.model().sourceModel()
+		model = self.tableView.model()
 
 		for i in xrange(model.rowCount()):
 			recordIds.append(model.index(i, enum.kRecordColumn_RecordId).data().toPyObject())
 
-		self.tagView.model().sourceModel().setFilter(recordIds)
+		self.tagView.model().sourceModel().setRecordFilter(recordIds)
+		self.tagView.resizeColumnsToContents()
+		self.displayRecordCount()
 
 
 	def addTag(self):
@@ -616,6 +603,8 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 		proxyModel.sourceModel().addRecordTags(tagId, self.selectedRecordIds())
 
 	def removeTag(self):
+		""" Delete a tag - ask for confirmation if tag is currently assigned to records
+		"""
 		proxyModel = self.tagView.model()
 		for proxyIndex in self.tagView.selectionModel().selectedRows():
 			assignedRecords = proxyModel.sourceModel().index(proxyModel.mapToSource(proxyIndex).row(), enum.kTagsColumn_RecordIds).data().toPyObject()
@@ -625,9 +614,9 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 						'There are %d records assigned to this tag\nSure you want to delete it?' % len(assignedRecords),
 						QtGui.QMessageBox.Yes|QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
 					continue
-			tagId = proxyModel.sourceModel().index(proxyModel.mapToSource(proxyIndex).row(), enum.kTagsColumn_TagId).data().toPyObject()
-			proxyModel.sourceModel().removeTag(tagId)
-
+			with utils.showWaitCursor():
+				tagId = proxyModel.sourceModel().index(proxyModel.mapToSource(proxyIndex).row(), enum.kTagsColumn_TagId).data().toPyObject()
+				proxyModel.sourceModel().removeTag(tagId)
 
 	def tagEditPopup(self, pos):
 		self.tableView.viewport().mapToGlobal(pos)
@@ -669,6 +658,7 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 		menu.addAction(action)
 		menu.exec_(self.tableView.viewport().mapToGlobal(pos))
 
+	@utils.showWaitCursorDecorator
 	def saveTagChanges(self, item):
 		""" Save tag changes to tag model
 		"""
@@ -684,12 +674,15 @@ class PydoshWindow(Ui_pydosh, QtGui.QMainWindow):
 			# If there's no selection available then close the tag menu
 			item.listWidget().parent().close()
 
-	def setTagFilter(self, tagIds):
-		""" Filter model by tags in response to tagView selection changed
-		"""
-		self.filterTagIds = tagIds
-		self.setFilter()
+	def scrollTo(self, text):
+		# Tell the model to highlight or un-highlight matching rows
+		self.tableView.model().sourceModel().highlightText(text)
 
+		if text:
+			currentIndex = self.tableView.model().index(0, enum.kRecordColumn_Description)
+			matches = self.tableView.model().match(currentIndex, QtCore.Qt.DisplayRole, text)
+			if matches:
+				self.tableView.scrollTo(matches[0], QtGui.QAbstractItemView.EnsureVisible)
 
 class TagListWidget(QtGui.QListWidget):
 	""" Simple extension to QListWidget to allow persistence of editor
