@@ -1,9 +1,13 @@
 from copy import deepcopy
+import operator
 from PyQt4 import QtCore, QtGui, QtSql
 import enum
 from database import db
+import utils
 import pydosh_rc
 
+import pdb
+
 class ImportException(Exception):
 	""" General exception for record import
 	"""
@@ -98,6 +102,16 @@ class ImportModel(QtCore.QAbstractTableModel):
 		self.__recordsRollback = deepcopy(self.__records)
 		self.__currentTimestamp = None
 
+	def flags(self, index):
+		""" Only allow selection on records that can be imported
+		"""
+		flags = super(ImportModel, self).flags(index)
+
+		if not self.__canImport(index):
+			return flags ^ QtCore.Qt.ItemIsSelectable
+		return flags
+
+
 	def loadRecords(self, records):
 		""" Import the records into our model
 			An input record is a tuple containing
@@ -131,7 +145,7 @@ class ImportModel(QtCore.QAbstractTableModel):
 		# Take a copy of the records in case we want to rollback
 		self.save()
 
-	def canImport(self, index):
+	def __canImport(self, index):
 		""" Returns True if the record at index can be imported,
 			ie no error and hasn't been imported yet
 		"""
@@ -141,19 +155,16 @@ class ImportModel(QtCore.QAbstractTableModel):
 	def rowCount(self, parent=QtCore.QModelIndex()):
 		return len(self.__records)
 
-	@property
 	def numBadRecords(self):
 		""" Returns the number of records with csv import errors
 		"""
 		return len([rec for rec in self.__records if not rec.valid])
 
-	@property
 	def numRecordsImported(self):
 		""" Returns the number of records that have already been imported
 		"""
 		return len([rec for rec in self.__records if rec.imported])
 
-	@property
 	def numRecordsToImport(self):
 		""" Returns the number of records left that can be imported
 		"""
@@ -172,21 +183,21 @@ class ImportModel(QtCore.QAbstractTableModel):
 			return QtCore.QVariant()
 
 		if role == QtCore.Qt.ToolTipRole:
-			return self.__records[item.row()].data
+			return QtCore.QVariant(self.__records[item.row()].data)
 
 		if role == QtCore.Qt.DisplayRole:
 			if item.column() == 0:
-				return self.__recordStatusToText(item)
+				return QtCore.QVariant(self.__recordStatusToText(item))
 			elif item.column() == 1:
-				return self.__records[item.row()].date
+				return QtCore.QVariant(self.__records[item.row()].date)
 			elif item.column() == 2:
-				return self.__records[item.row()].txdate
+				return QtCore.QVariant(self.__records[item.row()].txdate)
 			elif item.column() == 3:
-				return self.__records[item.row()].desc
+				return QtCore.QVariant(self.__records[item.row()].desc)
 			elif item.column() == 4:
-				return self.__records[item.row()].credit
+				return QtCore.QVariant(self.__records[item.row()].credit)
 			elif item.column() == 5:
-				return self.__records[item.row()].debit
+				return QtCore.QVariant(self.__records[item.row()].debit)
 
 		return QtCore.QVariant()
 
@@ -219,14 +230,16 @@ class ImportModel(QtCore.QAbstractTableModel):
 
 		if not rec.valid:
 			return rec.error
-		elif rec.imported:
+
+		if rec.imported:
 			return "Imported"
-		else:
-			return 'ready'
+
+		return 'ready'
 
 class RecordModel(QtSql.QSqlTableModel):
 	def __init__(self, parent=None):
 		super(RecordModel, self).__init__(parent=parent)
+		self._highlightText = None
 
 	def select(self):
 		status = super(RecordModel, self).select()
@@ -249,57 +262,35 @@ class RecordModel(QtSql.QSqlTableModel):
 		queryFilter = 'WHERE ' + queryFilter if queryFilter else ''
 
 		query = """
-			    SELECT r.recordid,
-					   r.checked,
-					   array_to_string(array_agg(t.tagname ORDER BY t.tagname), ','),
-					   r.checkdate,
-					   r.date,
-					   at.accountname,
-					   r.description,
-					   r.txdate,
-					   r.amount,
-					   r.insertdate,
-					   r.rawdata
-			      FROM records r
-			INNER JOIN accounttypes at ON at.accounttypeid=r.accounttypeid
-						AND r.userid=%(userid)s
-			 LEFT JOIN recordtags rt ON rt.recordid=r.recordid
-			 LEFT JOIN tags t ON rt.tagid=t.tagid
-					   %(filter)s
-			  GROUP BY r.recordid, at.accountname
-			  ORDER BY r.date, r.description, r.txdate, r.recordid
+                SELECT r.recordid,
+                       r.checked,
+                       array_to_string(array_agg(t.tagname ORDER BY t.tagname), ','),
+                       r.checkdate,
+                       r.date,
+                       r.accounttypeid,
+                       at.accountname,
+                       r.description,
+                       r.txdate,
+                       r.amount,
+                       r.insertdate,
+                       r.rawdata
+                  FROM records r
+            INNER JOIN accounttypes at ON at.accounttypeid=r.accounttypeid
+                   AND r.userid=%(userid)s
+             LEFT JOIN recordtags rt ON rt.recordid=r.recordid
+             LEFT JOIN tags t ON rt.tagid=t.tagid
+                       %(filter)s
+              GROUP BY r.recordid, at.accountname
+              ORDER BY r.date, r.recordid
 		""" % {'userid': db.userId, 'filter': queryFilter}
 
 		return query
 
-	def setItemsChecked(self, indexes):
-		""" Toggles the checked status on a list of model indexes
-			The checkdate column is set to the current timestamp is
-			checked, otherwise NULL is inserted
-
-			It's more efficient to make a bulk update query than
-			set every row via the model
-		"""
-		recordIds = [self.index(index.row(), enum.kRecordColumn_RecordId).data().toPyObject() for index in indexes]
-		query = QtSql.QSqlQuery("""
-			UPDATE records
-               SET checkdate = CASE WHEN checked=1 THEN NULL ELSE current_timestamp END,
-			       checked = CASE WHEN checked=1 THEN 0 ELSE 1 END
-			 WHERE recordid in (%s)
-			""" % ','.join(str(rec) for rec in recordIds))
-
-		if query.lastError().isValid():
-			return False
-
-		self.dataChanged.emit(indexes[0], indexes[-1])
-
-		return True
-
 	def deleteRecords(self, indexes):
 		recordIds = [self.index(index.row(), enum.kRecordColumn_RecordId).data().toPyObject() for index in indexes]
 		query = QtSql.QSqlQuery("""
-			DELETE FROM records
-			      WHERE recordid in (%s)
+            DELETE FROM records
+                  WHERE recordid in (%s)
 		""" % ','.join(str(rec) for rec in recordIds))
 
 		if query.lastError().isValid():
@@ -310,84 +301,121 @@ class RecordModel(QtSql.QSqlTableModel):
 
 		return True
 
+	def highlightText(self, text):
+		self._highlightText = text
+		self.reset()
+
 	def data(self, item, role=QtCore.Qt.DisplayRole):
 		""" Return data from the model, formatted for viewing
 		"""
 		if not item.isValid():
 			return QtCore.QVariant()
 
-		if role == QtCore.Qt.CheckStateRole and item.column() == enum.kRecordColumn_Checked:
-			if super(RecordModel, self).data(item).toBool():
-				return QtCore.Qt.Checked
-			else:
-				return QtCore.Qt.Unchecked
+		if role == QtCore.Qt.CheckStateRole:
+			if item.column() == enum.kRecordColumn_Checked:
+				if super(RecordModel, self).data(item).toBool():
+					return QtCore.QVariant(QtCore.Qt.Checked)
+				else:
+					return QtCore.QVariant(QtCore.Qt.Unchecked)
 
-		if role == QtCore.Qt.ToolTipRole:
+		elif role == QtCore.Qt.FontRole:
+			if item.column() == enum.kRecordColumn_Description:
+				if self._highlightText and item.data(QtCore.Qt.DisplayRole).toString().contains(self._highlightText, QtCore.Qt.CaseInsensitive):
+					font = QtGui.QFont()
+					font.setBold(True)
+					return QtCore.QVariant(font)
+
+		elif role == QtCore.Qt.ToolTipRole:
 			if item.column() == enum.kRecordColumn_Tags:
-				#Show tag names for this record
-				return self.record(item.row()).value(enum.kRecordColumn_Tags).toString()
+				# Show tag names for this record
+				return item.data(QtCore.Qt.UserRole)
 
 			elif item.column() == enum.kRecordColumn_Checked:
-				if super(RecordModel, self).data(self.index(item.row(), enum.kRecordColumn_Checked)).toBool():
-					return "Checked: " + super(RecordModel, self).data(
-							self.index(item.row(), enum.kRecordColumn_CheckDate)).toDateTime().toString("dd/MM/yy hh:mm")
+				if item.data(QtCore.Qt.CheckStateRole).toPyObject() == QtCore.Qt.Checked:
+					text = "Checked: " + super(RecordModel, self).data(
+						self.index(item.row(), enum.kRecordColumn_CheckDate)).toDateTime().toString("dd/MM/yy hh:mm")
+					return QtCore.QVariant(text)
 
 			elif item.column() == enum.kRecordColumn_Date:
-				return "Imported: " + super(RecordModel, self).data(
-							self.index(item.row(), enum.kRecordColumn_InsertDate)).toDateTime().toString("dd/MM/yy hh:mm")
+				# Show when the record was imported
+				text = "Imported: " + super(RecordModel, self).data(
+					self.index(item.row(), enum.kRecordColumn_InsertDate)).toDateTime().toString("dd/MM/yy hh:mm")
+				return QtCore.QVariant(text)
 
 			elif item.column() == enum.kRecordColumn_Description:
-				return super(RecordModel, self).data(item).toString()
+				# Full, raw text
+				return super(RecordModel, self).data(item, QtCore.Qt.DisplayRole)
 
-		if role == QtCore.Qt.UserRole and item.column() == enum.kRecordColumn_Tags:
-			# Return the number of tags
-			return super(RecordModel, self).data(item).toPyObject()
+		elif role == QtCore.Qt.UserRole:
+			if item.column() == enum.kRecordColumn_Tags:
+				# Tags as comma separated string (from database)
+				tags = super(RecordModel, self).data(item, QtCore.Qt.DisplayRole)
+				return tags if tags.toString() else QtCore.QVariant()
 
-		if role == QtCore.Qt.BackgroundColorRole:
+			elif item.column() == enum.kRecordColumn_Amount:
+				# Raw data - signed amount
+				return super(RecordModel, self).data(item, QtCore.Qt.DisplayRole)
+
+		elif role == QtCore.Qt.BackgroundColorRole:
 			# Indicate credit/debit with row colour
-			val, ok = self.record(item.row()).value(enum.kRecordColumn_Amount).toDouble()
-			if ok:
-				return QtGui.QColor("#b6ffac") if val > 0.0 else QtGui.QColor("#ffd3cf")
+			if self.index(item.row(), enum.kRecordColumn_Amount).data(QtCore.Qt.UserRole).toPyObject() > 0.0:
+				return QtCore.QVariant(QtGui.QColor("#b6ffac"))
+			else:
+				return QtCore.QVariant(QtGui.QColor("#ffd3cf"))
 
-		if role == QtCore.Qt.DecorationRole:
+		elif role == QtCore.Qt.DecorationRole:
 			if item.column() == enum.kRecordColumn_Tags:
 				# Show tag icon if we have any
-				if super(RecordModel, self).data(item).toPyObject():
-					return QtGui.QIcon(':/icons/tag_yellow.png')
+				if item.data(QtCore.Qt.UserRole).toString():
+					return QtCore.QVariant(QtGui.QIcon(':/icons/tag_yellow.png'))
 
-		if role == QtCore.Qt.DisplayRole:
+		elif role == QtCore.Qt.DisplayRole:
 			if item.column() in (enum.kRecordColumn_Checked, enum.kRecordColumn_Tags):
 				# Don't display anything for these fields
 				return QtCore.QVariant()
 
 			elif item.column() == enum.kRecordColumn_Amount:
 				# Display absolute currency values. credit/debit is indicated by background colour
-				val, ok = super(RecordModel, self).data(item).toDouble()
-				if ok:
-					return QtCore.QString.number(abs(val), 'f', 2)
+				return QtCore.QVariant('%.02f' % abs(super(RecordModel, self).data(item).toPyObject()))
 
 			elif item.column() == enum.kRecordColumn_Description:
 				# Replace multiple spaces with single
-				return super(RecordModel, self).data(item).toString().replace(QtCore.QRegExp('[ ]+'), ' ')
-
-			elif item.column() == enum.kRecordColumn_Date:
-				return super(RecordModel, self).data(item).toDate()
+				val = super(RecordModel, self).data(item).toString().replace(QtCore.QRegExp('[ ]+'), ' ')
+				return QtCore.QVariant(val)
 
 			elif item.column() == enum.kRecordColumn_Txdate:
+				# Reformat date to only display time if available
 				if super(RecordModel, self).data(item).toDateTime().time().toString() == "00:00:00":
-					return super(RecordModel, self).data(item).toDate()
-				else:
-					return super(RecordModel, self).data(item).toDateTime()
+					return QtCore.QVariant(super(RecordModel, self).data(item).toDate())
 
 		return super(RecordModel, self).data(item, role)
 
 	def setData(self, index, value, role=QtCore.Qt.EditRole):
 		""" Save new checkstate role changes in database
+		
+			Only send a single dataChanged signal for the whole row, not two separate
 		"""
+		status = False
 		if role == QtCore.Qt.CheckStateRole and index.column() == enum.kRecordColumn_Checked:
-			return self.setItemsChecked([index])
+			with utils.signalsBlocked(self):
+				if value.toPyObject() == QtCore.Qt.Checked:
+					status = (
+						super(RecordModel, self).setData(index, 1, QtCore.Qt.EditRole) and 
+						super(RecordModel, self).setData(self.index(index.row(), enum.kRecordColumn_CheckDate),
+														QtCore.QDateTime.currentDateTimeUtc(), 
+														QtCore.Qt.EditRole)
+					)
+				elif value.toPyObject() == QtCore.Qt.Unchecked:
+					status = (
+						super(RecordModel, self).setData(index, 0, QtCore.Qt.EditRole) and
+						super(RecordModel, self).setData(self.index(index.row(), enum.kRecordColumn_CheckDate),
+														QtCore.QVariant(), 
+														QtCore.Qt.EditRole)
+					)
+		if status:
+			self.dataChanged.emit(index, index)
 
-		return False
+		return status
 
 	def headerData (self, section, orientation, role=QtCore.Qt.DisplayRole):
 		""" Set the header labels for the view
@@ -399,7 +427,7 @@ class RecordModel(QtSql.QSqlTableModel):
 				return "Tags"
 			elif section == enum.kRecordColumn_Date:
 				return "Date"
-			elif section == enum.kRecordColumn_AccountType:
+			elif section == enum.kRecordColumn_AccountTypeName:
 				return "Account"
 			elif section == enum.kRecordColumn_Description:
 				return "Description"
@@ -416,14 +444,18 @@ class TagModel(QtSql.QSqlTableModel):
 
 	def __init__(self, parent=None):
 		super(TagModel, self).__init__(parent=parent)
-		self.__selectedTagIds = set()
+		self.__selectedTagNames = set()
 
 		self.setTable('tags')
 		self.setEditStrategy(QtSql.QSqlTableModel.OnFieldChange)
 		super(TagModel, self).select()
 
-	def setFilter(self, recordIds):
-		super(TagModel, self).setFilter(','.join(str(rec) for rec in recordIds))
+	def setRecordFilter(self, recordIds):
+		""" List of record ids to limit tag data to display
+			If no record ids are given then we still need to set 
+			"0" to ensure that no record ids are matched
+		"""
+		self.setFilter(','.join([str(rec) for rec in recordIds or [0]]))
 
 	def clearSelection(self):
 		for row in xrange(self.rowCount()):
@@ -433,22 +465,28 @@ class TagModel(QtSql.QSqlTableModel):
 	def setData(self, index, value, role=QtCore.Qt.EditRole):
 		""" Handle checkstate role changes
 		"""
-		if role == QtCore.Qt.CheckStateRole and index.column() == enum.kTagsColumn_TagName:
-			tagId = self.index(index.row(), enum.kTagsColumn_TagId).data().toPyObject()
-
-			if value.toPyObject() == QtCore.Qt.Checked:
-				if tagId in self.__selectedTagIds:
-					return False
-				self.__selectedTagIds.add(tagId)
-			else:
-				if tagId not in self.__selectedTagIds:
-					return False
-				
-				self.__selectedTagIds.remove(tagId)
-
-			self.dataChanged.emit(index, index)
-			self.selectionChanged.emit(self.__selectedTagIds)
-			return True 
+		if index.column() == enum.kTagsColumn_TagName:
+			if role == QtCore.Qt.CheckStateRole:
+				tagName = index.data().toPyObject()
+	
+				if value.toPyObject() == QtCore.Qt.Checked:
+					if tagName in self.__selectedTagNames:
+						# Do nothing if tag has not changed
+						return False
+					self.__selectedTagNames.add(tagName)
+				else:
+					if tagName not in self.__selectedTagNames:
+						# Do nothing if tag has not changed
+						return False
+					self.__selectedTagNames.remove(tagName)
+	
+				self.dataChanged.emit(index, index)
+				self.selectionChanged.emit(self.__selectedTagNames)
+				return True 
+	
+			elif role == QtCore.Qt.EditRole:
+				# Save changes to tag name in database
+				return super(TagModel, self).setData(index, value, role)
 
 		return False
 
@@ -457,7 +495,8 @@ class TagModel(QtSql.QSqlTableModel):
 		if role == QtCore.Qt.DisplayRole:
 
 			if item.column() == enum.kTagsColumn_RecordIds:
-				return set([int(i) for i in super(TagModel, self).data(item).toString().split(',') if i])
+				tags = set([int(i) for i in super(TagModel, self).data(item).toString().split(',') if i])
+				return QtCore.QVariant(tags)
 
 			elif item.column() in (enum.kTagsColumn_Amount_in, enum.kTagsColumn_Amount_out):
 				amount = '%.2f' % super(TagModel, self).data(item).toPyObject()
@@ -467,18 +506,23 @@ class TagModel(QtSql.QSqlTableModel):
 					return QtCore.QVariant(amount)
 
 		if  role == QtCore.Qt.CheckStateRole and item.column() == enum.kTagsColumn_TagName:
-			tagId = self.index(item.row(), enum.kTagsColumn_TagId).data().toPyObject()
-			if tagId in self.__selectedTagIds:
-				return QtCore.Qt.Checked
+			if item.data().toPyObject() in self.__selectedTagNames:
+				return QtCore.QVariant(QtCore.Qt.Checked)
 			else:
-				return QtCore.Qt.Unchecked
+				return QtCore.QVariant(QtCore.Qt.Unchecked)
 
 		return super(TagModel, self).data(item, role)
 
 	def flags(self, index):
 		flags = super(TagModel, self).flags(index)
+
 		if index.column() == enum.kTagsColumn_TagName:
 			flags |= QtCore.Qt.ItemIsUserCheckable
+		
+		# Only allow tag name to be editable
+		if index.column() != enum.kTagsColumn_TagName:
+			flags ^= QtCore.Qt.ItemIsEditable
+
 		return flags
 
 	def selectStatement(self):
@@ -524,7 +568,7 @@ class TagModel(QtSql.QSqlTableModel):
 		query.addBindValue(db.userId)
 
 		if not query.exec_():
-			raise query.lastError().text()
+			raise Exception(query.lastError().text())
 		
 		query.next()
 		insertId = query.value(0).toPyObject() 
@@ -550,7 +594,7 @@ class TagModel(QtSql.QSqlTableModel):
 	def addRecordTags(self, tagId, recordIds):
 		if not recordIds:
 			return False
-		
+
 		# Remove records that already have this tag
 		currentIndex = self.index(0, enum.kTagsColumn_TagId)
 		match = self.match(currentIndex, QtCore.Qt.DisplayRole, tagId, 1, QtCore.Qt.MatchExactly)
@@ -591,48 +635,238 @@ class TagModel(QtSql.QSqlTableModel):
 		self.tagsChanged.emit()
 		return self.select()
 
+class TagProxyModel(QtGui.QSortFilterProxyModel):
+	# Signal emitted whenever there is a change to the filter
+	filterChanged = QtCore.pyqtSignal()
 
-class SortProxyModel(QtGui.QSortFilterProxyModel):
 	def __init__(self, parent=None):
-		super(SortProxyModel, self).__init__(parent=parent)
+		super(TagProxyModel, self).__init__(parent=parent)
 
 	def lessThan(self, left, right):
 		""" Define the comparison to ensure column data is sorted correctly
 		"""
+		if left.column() in (enum.kTagsColumn_Amount_in, enum.kTagsColumn_Amount_out):
+			return left.data().toDouble()[0] > right.data().toDouble()[0]
+
+		return super(TagProxyModel, self).lessThan(left, right)
+
+class RecordProxyModel(QtGui.QSortFilterProxyModel):
+	# Signal emitted whenever there is a change to the filter
+	filterChanged = QtCore.pyqtSignal()
+
+	def __init__(self, parent=None):
+		super(RecordProxyModel, self).__init__(parent=parent)
+		self.__reset()
+
+	def __reset(self):
+		""" Create or re-create all filter
+		"""
+		self._startDate = None
+		self._endDate = None
+		self._insertDate = None
+		self._accountids = None
+		self._hasTags = None
+		self._checked = None
+		self._creditFilter = None
+		self._description = None
+		self._amountFilter = None
+		self._tagFilter = None
+
+	def clearFilters(self):
+		""" Clears all filters - does *not* call invalidate
+		"""
+		self.__reset()
+
+	def setInsertDate(self, insertDate):
+		self._startDate = None
+		self._endDate = None
+		self._insertDate = insertDate
+		self.invalidateFilter()
+
+	def setStartDate(self, startDate):
+		self._insertDate = None
+		self._startDate = startDate
+		self.invalidateFilter()
+
+	def setEndDate(self, date):
+		self._insertDate = None
+		self._endDate = date
+		self.invalidateFilter()
+
+	def setAccountFilter(self, accountIds):
+		if accountIds != self._accountids:
+			self._accountids = accountIds
+			self.invalidateFilter()
+			
+	def setHasTagsFilter(self, value):
+		""" Set basic tag filter
+
+			selection:
+				None  - no filter
+				True  - filter with tags
+				False - filter with no tags
+		"""
+		if value != self._hasTags:
+			self._hasTags = value
+			self.invalidateFilter()
+
+	def setTagFilter(self, tags):
+		if tags != self._tagFilter:
+			self._tagFilter = tags.copy()
+			self.invalidateFilter()
+
+	def setCheckedFilter(self, value):
+		""" Checked records filter
+
+			selection:
+				None  - all
+				True  - filter only checked
+				False - filter not checked
+		"""
+		if value != self._checked:
+			self._checked = value
+			self.invalidateFilter()
+	
+	def setCreditFilter(self, value):
+		""" Credit amount filter
+
+			selection:
+				None  - all
+				True  - filter on credit
+				False - filter on debit
+		"""
+		if value != self._creditFilter:
+			self._creditFilter = value
+			self.invalidateFilter()
+
+	def setDescriptionFilter(self, text):
+		""" Filter by description (case insensitive)
+		"""
+		if text != self._description:
+			self._description = text
+			self.invalidateFilter()
+
+	def setAmountFilter(self, text, op=None):
+		""" Set amount filter with optional operator
+			If operator is None then a string comparison is done on amount start
+		""" 
+		if text != self._amountFilter or op != self._amountOperator:
+			self._amountFilter = text
+			self._amountOperator = op
+			self.invalidateFilter()
+
+	def invalidateFilter(self):
+		""" Override invalidateFilter so that we can emit the filterChanged signal
+		"""
+		super(RecordProxyModel, self).invalidateFilter()
+		self.filterChanged.emit()
+
+	def filterAcceptsRow(self, sourceRow, parent):
+		if self._startDate:
+			if self.sourceModel().index(sourceRow, enum.kRecordColumn_Date, parent).data().toDate() < self._startDate:
+				return False
+
+		if self._endDate:
+			if self.sourceModel().index(sourceRow, enum.kRecordColumn_Date, parent).data().toDate() > self._endDate:
+				return False
+
+		if self._insertDate:
+			if self.sourceModel().index(sourceRow, enum.kRecordColumn_InsertDate, parent).data().toDateTime() != self._insertDate:
+				return False
+
+		if self._accountids:
+			if self.sourceModel().index(sourceRow, enum.kRecordColumn_AccountTypeId).data().toPyObject() not in self._accountids:
+				return False
+
+		if self._hasTags is not None:
+			hasTags = self.sourceModel().index(sourceRow, enum.kRecordColumn_Tags).data(QtCore.Qt.UserRole).toBool()
+			if self._hasTags != hasTags:
+				return False
+
+		if self._checked is not None:
+			isChecked = self.sourceModel().index(sourceRow, enum.kRecordColumn_Checked).data(QtCore.Qt.CheckStateRole).toPyObject() == QtCore.Qt.Checked
+			if self._checked != isChecked:
+				return False
+
+		if self._creditFilter is not None:
+			amount = self.sourceModel().index(sourceRow, enum.kRecordColumn_Amount).data(QtCore.Qt.UserRole).toPyObject()
+			if self._creditFilter != (amount >= 0.0):
+				return False
+
+		if self._description:
+			description = self.sourceModel().index(sourceRow, enum.kRecordColumn_Description).data().toString()
+			if not description.contains(self._description, QtCore.Qt.CaseInsensitive):
+				return False
+
+		if self._amountFilter:
+			amount = self.sourceModel().index(sourceRow, enum.kRecordColumn_Amount).data().toString()
+			if self._amountOperator is None:
+				# Filter as string matching start
+				if not amount.startsWith(self._amountFilter):
+					return False
+			else:
+				# Use operator to perform match
+				if not self._amountOperator(float(amount), float(self._amountFilter)):
+					return False
+				
+		if self._tagFilter:
+			tags = self.sourceModel().index(sourceRow, enum.kRecordColumn_Tags).data(QtCore.Qt.UserRole).toString()
+			if not set(self._tagFilter).intersection(set(tags.split(','))):
+				return False
+
+		return True
+
+	def lessThan(self, left, right):
+		""" Define the comparison to ensure column data is sorted correctly
+		"""
+		leftVal = None
+		rightVal = None
+
 		if left.column() == enum.kRecordColumn_Tags:
-			return self.sourceModel().data(left, QtCore.Qt.UserRole) < self.sourceModel().data(right, QtCore.Qt.UserRole)
+			leftVal = len(left.data(QtCore.Qt.UserRole).toString().split(',', QtCore.QString.SkipEmptyParts))
+			rightVal = len(right.data(QtCore.Qt.UserRole).toString().split(',', QtCore.QString.SkipEmptyParts))
 
 		elif left.column() == enum.kRecordColumn_Checked:
-			return self.sourceModel().data(left, QtCore.Qt.CheckStateRole) < self.sourceModel().data(right, QtCore.Qt.CheckStateRole)
+			leftVal = left.data(QtCore.Qt.CheckStateRole).toPyObject()
+			rightVal = right.data(QtCore.Qt.CheckStateRole).toPyObject()
 
 		elif left.column() == enum.kRecordColumn_Amount:
-			leftVal, leftOk = self.sourceModel().record(left.row()).value(enum.kRecordColumn_Amount).toDouble()
-			rightVal, rightOk = self.sourceModel().record(right.row()).value(enum.kRecordColumn_Amount).toDouble()
-			if leftOk and rightOk:
-				return leftVal < rightVal
+			leftVal,_ = left.data(QtCore.Qt.UserRole).toDouble()
+			rightVal,_ = right.data(QtCore.Qt.UserRole).toDouble()
 
-		return super(SortProxyModel, self).lessThan(left, right)
+		elif left.column() == enum.kRecordColumn_Date:
+			leftVal =  left.data().toDate()
+			rightVal = right.data().toDate()
 
+			if leftVal == rightVal:
+				# Dates are the same - sort by recordId just to ensure the results are consistent
+				leftVal = self.sourceModel().index(left.row(), enum.kRecordColumn_RecordId).data().toPyObject()
+				rightVal = self.sourceModel().index(right.row(), enum.kRecordColumn_RecordId).data().toPyObject()
 
-class AccountModel(QtSql.QSqlTableModel):
+		if leftVal or rightVal:
+			return leftVal < rightVal
+
+		return super(RecordProxyModel, self).lessThan(left, right)
+
+class AccountEditModel(QtSql.QSqlTableModel):
 	def __init__(self, parent=None):
-		super(AccountModel, self).__init__(parent=parent)
+		super(AccountEditModel, self).__init__(parent=parent)
 
 	def data(self, item, role=QtCore.Qt.DisplayRole):
 		if not item.isValid():
 			return QtCore.QVariant()
 
 		if role == QtCore.Qt.BackgroundColorRole and self.isDirty(item):
-				return QtGui.QColor(255,156,126)
+				return QtCore.QVariant(QtGui.QColor(255,156,126))
 
-		return super(AccountModel, self).data(item, role)
+		return super(AccountEditModel, self).data(item, role)
 
 	def setData(self, index, value, role=QtCore.Qt.EditRole):
 		# Don't flag cell as changed when it hasn't
-		if role == QtCore.Qt.EditRole and self.data(index, role) == value:
+		if role == QtCore.Qt.EditRole and index.data(QtCore.Qt.DisplayRole) == value:
 			return False
 
-		return super(AccountModel, self).setData(index, value, role)
+		return super(AccountEditModel, self).setData(index, value, role)
 
 	def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
 
@@ -654,45 +888,50 @@ class AccountModel(QtSql.QSqlTableModel):
 
 		return QtCore.QVariant()
 
-class CheckComboModel(QtSql.QSqlTableModel):
-
+class AccountsModel(QtSql.QSqlTableModel):
+	""" Model for accounts which stores selection.
+		Underlying model is QSqlTableModel
+	"""
 	def __init__(self, parent=None):
-		super(CheckComboModel, self).__init__(parent=parent)
+		super(AccountsModel, self).__init__(parent=parent)
 		self._checkedItems = set()
-		self.__userRoleColumn = None
-
-	def setUserRoleColumn(self, column):
-		""" Set the database table column to use for UserRole data
-		"""
-		self.__userRoleColumn = column
 
 	def flags(self, index):
-		return super(CheckComboModel, self).flags(index) | QtCore.Qt.ItemIsUserCheckable
+		""" Enable checkable account names
+		"""
+		flags = super(AccountsModel, self).flags(index)
+
+		if index.column() == enum.kAccountTypeColumn_AccountName:
+			flags |= QtCore.Qt.ItemIsUserCheckable
+
+		return flags
 
 	def data(self, index, role):
 		if not index.isValid():
 			return QtCore.QVariant()
 
 		if role == QtCore.Qt.CheckStateRole:
-			if index.row() in self._checkedItems:
-				return QtCore.Qt.Checked
-			else:
-				return QtCore.Qt.Unchecked
+			if index.column() == enum.kAccountTypeColumn_AccountName:
+				accountName = self.index(index.row(), enum.kAccountTypeColumn_AccountName).data().toString()
+				if accountName in self._checkedItems:
+					return QtCore.QVariant(QtCore.Qt.Checked)
+				else:
+					return QtCore.QVariant(QtCore.Qt.Unchecked)
 
-		if role == QtCore.Qt.UserRole and self.__userRoleColumn is not None:
-			return self.record(index.row()).value(self.__userRoleColumn).toPyObject()
-
-		return super(CheckComboModel, self).data(index, role)
+		return super(AccountsModel, self).data(index, role)
 
 	def setData(self, index, value, role):
+		""" Set check state and preserve account name for use in data()
+		"""
 		if role == QtCore.Qt.CheckStateRole:
-
-			if value.toPyObject() == QtCore.Qt.Checked:
-				self._checkedItems.add(index.row())
-			else:
-				self._checkedItems.remove(index.row())
-
-			self.dataChanged.emit(index, index)
-			return True
+			if index.column() == enum.kAccountTypeColumn_AccountName:
+				accountName = self.index(index.row(), enum.kAccountTypeColumn_AccountName).data().toString()
+				if value.toPyObject() == QtCore.Qt.Checked:
+					self._checkedItems.add(accountName)
+				else:
+					self._checkedItems.remove(accountName)
+	
+				self.dataChanged.emit(index, index)
+				return True
 
 		return False
